import requests
from lxml import objectify
from lxml.etree import ElementBase


def DumpNode(node):
    d = {}
    for key, value in vars(node).items():
        if isinstance(value, ElementBase):
            value = DumpNode(value)
        d[key] = value
    if len(d) == 0:
        return str(node)
    return d


class NWSCAPParser:
    def __init__(self, url=None, rawXML=None):
        self.rawXML = rawXML
        self.url = url

        if self.rawXML is None and self.url is None:
            raise Exception('You must instantiate this object with either "url" or "rawXML"')

        self.entries = {
            # str(id): CAPEntryObj,
        }

        self.Update()

    def Update(self):
        xmlString = self.rawXML or requests.get(self.url).text
        # print('xmlString=', xmlString)
        xmlObj = objectify.fromstring(xmlString.encode())

        for entryTag in xmlObj.entry:
            entryObj = CAPEntry(xmlObj, entryTag)
            if entryObj.id not in self.entries:
                self.entries[entryObj.id] = entryObj


class CAPEntry:
    def __init__(self, xmlObj, entry):
        self.xmlObj = xmlObj
        self.entry = entry

    def __getattr__(self, item):
        try:
            res = getattr(self.entry, item)
            return res
        except:
            pass

        res = self.entry.find(f'{item}')
        if res:
            return res
        else:
            return self.entry.find(f'{{{self.xmlObj.nsmap["cap"]}}}{item}')

    def __iter__(self):
        for key, value in DumpNode(self).items():
            yield key, value

    def dict(self):
        ret = {}
        for key in  [
            'id',
            'title',
            'severity',
            'urgency',
            'certainty',
            'status',
            'msgType',
            'category',
            'areaDesc',
        ]:
            ret[key] = str(getattr(self, key))
        return ret

    def __str__(self):
        return f'<CAPEntry: {DumpNode(self.entry)}>'


if __name__ == '__main__':
    MY_STATE = 'NC'
    cap = NWSCAPParser('http://alerts.weather.gov/cap/{}.php?x=1'.format(MY_STATE))
    import json
    for ID, entry in cap.entries.items():
        print('################################')
        print('ID=', ID)
        print('entry.id=', entry.id)
        print('entry=', entry)
        print('entry.title=', entry.title)

        print('entry.severity=', entry.severity)
        print('entry.urgency=', entry.urgency)
        print('entry.certainty=', entry.certainty)

        print('entry.status=', entry.status)
        print('entry.msgType=', entry.msgType)
        print('entry.category=', entry.category)
        print('entry.areaDesc=', entry.areaDesc)

        print('dict()=', entry.dict())
        print('json()=', json.dumps(entry.dict()))


## {{{ http://code.activestate.com/recipes/534109/ (r8)
import re
import xml.sax.handler
import dateutil.parser as duparser
import calendar

# used to clean text by removing multiple spaces
R1 = re.compile(r"^\s{2,}", re.MULTILINE)

def clean_text(field_name, raw_text):
    if field_name in ['description','instruction']:
        return R1.sub(" ", raw_text.strip()).replace('\n',' ')
    return raw_text

def convert_timestamp(t):
    #e.g., 2012-03-15T22:18:00-04:00
    dt = duparser.parse(t)
    return calendar.timegm(dt.utctimetuple())

def xml2obj(src):
    """
    A simple function to converts XML data into native Python object.
    """

    non_id_char = re.compile('[^_0-9a-zA-Z]')
    def _name_mangle(name):
        return non_id_char.sub('_', name)

    class DataNode(object):
        def __init__(self):
            self._attrs = {}    # XML attributes and child elements
            self.data = None    # child text data
        def __len__(self):
            # treat single element as a list of 1
            return 1
        def __getitem__(self, key):
            if isinstance(key, basestring):
                return self._attrs.get(key,None)
            else:
                return [self][key]
        def __contains__(self, name):
            return name in self._attrs
        def __nonzero__(self):
            return bool(self._attrs or self.data)
        def __getattr__(self, name):
            if name.startswith('__'):
                # need to do this for Python special methods???
                raise AttributeError(name)
            return self._attrs.get(name,None)
        def _add_xml_attr(self, name, value):
            if name in self._attrs:
                # multiple attribute of the same name are represented by a list
                children = self._attrs[name]
                if not isinstance(children, list):
                    children = [children]
                    self._attrs[name] = children
                children.append(value)
            else:
                self._attrs[name] = value
        def __str__(self):
            return self.data or ''
        def __repr__(self):
            items = sorted(self._attrs.items())
            if self.data:
                items.append(('data', self.data))
            return u'{%s}' % ', '.join([u"'%s':%s" % (k,repr(v)) for k,v in items])

    class TreeBuilder(xml.sax.handler.ContentHandler):
        def __init__(self):
            self.stack = []
            self.root = DataNode()
            self.current = self.root
            self.text_parts = []
        def startElement(self, name, attrs):
            self.stack.append((self.current, self.text_parts))
            self.current = DataNode()
            self.text_parts = []
            # xml attributes --> python attributes
            for k, v in attrs.items():
                self.current._add_xml_attr(_name_mangle(k), v)
        def endElement(self, name):
            text = ''.join(self.text_parts).strip()
            if text:
                self.current.data = clean_text(name,text)
            if self.current._attrs:
                obj = self.current
            else:
                # a text only node is simply represented by the string
                obj = clean_text(name,text) or ''
            self.current, self.text_parts = self.stack.pop()
            self.current._add_xml_attr(_name_mangle(name), obj)
            try:
                if str(obj).strip() != '':
                    obj_epoch = convert_timestamp(obj)
                    self.current._add_xml_attr(_name_mangle("%s_epoch"%name), obj_epoch)
            except:
                pass
        def characters(self, content):
            self.text_parts.append(content)


    builder = TreeBuilder()
    if isinstance(src,basestring):
        xml.sax.parseString(src, builder)
    else:
        xml.sax.parse(src, builder)
    return builder.root._attrs.values()[0]
## end of http://code.activestate.com/recipes/534109/ }}}



## {{{ http://code.activestate.com/recipes/577305/ (r1)
states = {
        'AK': 'Alaska',
        'AL': 'Alabama',
        'AR': 'Arkansas',
        'AS': 'American Samoa',
        'AZ': 'Arizona',
        'CA': 'California',
        'CO': 'Colorado',
        'CT': 'Connecticut',
        'DC': 'District of Columbia',
        'DE': 'Delaware',
        'FL': 'Florida',
        'GA': 'Georgia',
        'GU': 'Guam',
        'HI': 'Hawaii',
        'IA': 'Iowa',
        'ID': 'Idaho',
        'IL': 'Illinois',
        'IN': 'Indiana',
        'KS': 'Kansas',
        'KY': 'Kentucky',
        'LA': 'Louisiana',
        'MA': 'Massachusetts',
        'MD': 'Maryland',
        'ME': 'Maine',
        'MI': 'Michigan',
        'MN': 'Minnesota',
        'MO': 'Missouri',
        'MP': 'Northern Mariana Islands',
        'MS': 'Mississippi',
        'MT': 'Montana',
        'NA': 'National',
        'NC': 'North Carolina',
        'ND': 'North Dakota',
        'NE': 'Nebraska',
        'NH': 'New Hampshire',
        'NJ': 'New Jersey',
        'NM': 'New Mexico',
        'NV': 'Nevada',
        'NY': 'New York',
        'OH': 'Ohio',
        'OK': 'Oklahoma',
        'OR': 'Oregon',
        'PA': 'Pennsylvania',
        'PR': 'Puerto Rico',
        'RI': 'Rhode Island',
        'SC': 'South Carolina',
        'SD': 'South Dakota',
        'TN': 'Tennessee',
        'TX': 'Texas',
        'UT': 'Utah',
        'VA': 'Virginia',
        'VI': 'Virgin Islands',
        'VT': 'Vermont',
        'WA': 'Washington',
        'WI': 'Wisconsin',
        'WV': 'West Virginia',
        'WY': 'Wyoming'
}
## end of http://code.activestate.com/recipes/577305/ }}}



import sys
from pprint import pprint

from nwscapparser3 import NWSCAPParser


def test_basic_fields(alert):
    print('---- basic fields ----')
    print(alert.identifier)
    print(alert.info.effective)
    print(f'{len(alert.FIPS6)} FIPS6 codes:{alert.FIPS6}')
    print(f'{len(alert.UGC)} UGC codes:{alert.UGC}')
    print(f'{len(alert.INFO_PARAMS)} INFO_PARAMS:{alert.INFO_PARAMS}')


def test_dict_dump(alert):
    print('---- dict dump ----')
    pprint(alert.as_dict())


def test_json_dump(alert):
    print('---- json dump ----')
    print(alert.as_json())


if __name__ == '__main__':
    # first command line arg is assumed to be a full URL to a CAP
    import requests

    def TestLocalXML():
        print('--Testing parse of a local XML file---')

        filepath = r'cap.2020-05-07T14_12_00.xml'
        with open(filepath, 'rt') as file:
            src = file.read()

        cap = NWSCAPParser(rawXML=src)

        for ID, entry in cap.entries.items():
            print('ID=', ID, ', entry=', entry)

    def TestRemoteURL():
        print('--Testing parse of a remote file---')

        cap = NWSCAPParser(url='http://alerts.weather.gov/cap/nc.php?x=1')

        for ID, entry in cap.entries.items():
            print('ID=', ID, ', entry=', entry)

    TestLocalXML()
    TestRemoteURL()


from distutils.core import setup
import glob
from setuptools import setup
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'readme.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='nwscapparser3',
    version='3.0.6',
    # 3.0.6 - added CAPEntry.dict() so that items can be easily json-ized

    description='NWS CAP Parser for Python3',
    long_description=long_description,
    author='Grant Miller. Based on the work of Robert Morris (robert@emthree.com)',
    author_email='grant@grant-miller.com',
    url='https://github.com/GrantGMiller/NWS-CAP-parser',
    packages=['nwscapparser3'],
    license="PSF",
    keywords="python3 cap common alert protocol grant miller robert morris xml",
    install_requires=['requests', 'lxml'],
    data_files=[
        ('', glob.glob('*.xml')),
        ('', ['demo.py', 'readme.rst', 'LICENSE'])]
)

# to push to PyPI

# python -m setup.py sdist bdist_wheel
# twine upload dist/*
